package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import edu.wpi.first.math.estimator.PoseEstimator;
import edu.wpi.first.math.geometry.Pose2d;
import edu.wpi.first.math.geometry.Rotation2d;
import edu.wpi.first.math.geometry.Transform2d;
import edu.wpi.first.math.geometry.Translation2d;
import edu.wpi.first.math.interpolation.TimeInterpolatableBuffer;
import edu.wpi.first.math.util.Units;
import edu.wpi.first.networktables.BooleanSubscriber;
import edu.wpi.first.networktables.DoubleSubscriber;
import edu.wpi.first.networktables.NetworkTable;
import edu.wpi.first.networktables.NetworkTableInstance;
import edu.wpi.first.wpilibj.Timer;
import frc.robot.constants.VisionConstants;

public class Vision{

    private DoubleSubscriber blackRobotRelativeX;
    private DoubleSubscriber blackRobotRelativeY;
    private DoubleSubscriber blackRobotRelativeZ;

    private DoubleSubscriber blackClosestTag;
    private BooleanSubscriber blackTagDetected;

    private DoubleSubscriber blackFramerate;

    private DoubleSubscriber orange_tx;
    private DoubleSubscriber orange_ty;
    private DoubleSubscriber orange_dist;

    private DoubleSubscriber orangeClosestTag;
    private BooleanSubscriber orangeTagDetected;

    private DoubleSubscriber orangeFramerate;

    private double[] orangeCamPose = {0,0,0,0,0};
    private double[] blackCamPose = {0,0,0,0,0};

    public Vision(){
        NetworkTableInstance inst = NetworkTableInstance.getDefault();

        NetworkTable blackVisionTable = inst.getTable("black_Fiducial");
        NetworkTable orangeVisionTable = inst.getTable("orange_Fiducial");

        blackRobotRelativeX = blackVisionTable.getDoubleTopic("blackRelativeX").subscribe(0.0);
        blackRobotRelativeY = blackVisionTable.getDoubleTopic("blackRelativeY").subscribe(0.0);
        blackRobotRelativeZ = blackVisionTable.getDoubleTopic("blackRelativeZ").subscribe(0.0);

        blackClosestTag = blackVisionTable.getDoubleTopic("blackClosestTag").subscribe(0.0);
        blackTagDetected = blackVisionTable.getBooleanTopic("blackTagDetected").subscribe(false);

        blackFramerate = blackVisionTable.getDoubleTopic("blackFPS").subscribe(0.0);

        orange_tx = orangeVisionTable.getDoubleTopic("tx").subscribe(0.0);
        orange_ty = orangeVisionTable.getDoubleTopic("ty").subscribe(0.0);
        orange_dist = orangeVisionTable.getDoubleTopic("totalDist").subscribe(0.0);

        orangeClosestTag = orangeVisionTable.getDoubleTopic("orangeClosestTag").subscribe(0.0);
        orangeTagDetected = orangeVisionTable.getBooleanTopic("orangeTagDetected").subscribe(false);

        orangeFramerate = orangeVisionTable.getDoubleTopic("orangeFPS").subscribe(0.0);
    
    }
    
    public Pose2d relativeToGlobalPose2d(int tagID, Translation2d relativeCoords, double timestamp, TimeInterpolatableBuffer<Double> gyroBuffer){
        Pose2d tag2dPose = new Pose2d(VisionConstants.globalTagCoords[tagID][0],
            VisionConstants.globalTagCoords[tagID][1],
            new Rotation2d());

        Pose2d relative = new Pose2d(relativeCoords, new Rotation2d(gyroBuffer.getSample(timestamp).get()));

        Transform2d relative2dTransformation = new Transform2d(relative.getTranslation(), relative.getRotation());

        Pose2d globalPose = tag2dPose.transformBy(relative2dTransformation.inverse());

        return new Pose2d(globalPose.getTranslation(), new Rotation2d(gyroBuffer.getSample(timestamp).get()));
    }

    public Pose2d CameraToGlobalPose2d(int tagID, double totalDist, double tx, double ty, double timestamp, TimeInterpolatableBuffer<Double> gyroBuffer){

        System.out.println(gyroBuffer.getSample(timestamp));

        double distance2d = totalDist * Math.cos(-orangeCamPose[1] - ty);

        Rotation2d camToTagRotation = new Rotation2d(gyroBuffer.getSample(timestamp).get()).plus(new Rotation2d(orangeCamPose[0] - tx));

        Pose2d tagPose2d = new Pose2d(VisionConstants.globalTagCoords[tagID][0],
            VisionConstants.globalTagCoords[tagID][1],
            new Rotation2d());

        Translation2d fieldToCameraTranslation = new Pose2d(tagPose2d.getTranslation(), camToTagRotation.plus(Rotation2d.kPi))
            .transformBy(new Transform2d(distance2d, 0.0, new Rotation2d()))
            .getTranslation();
        Pose2d robotPose = new Pose2d(fieldToCameraTranslation,
            new Rotation2d(gyroBuffer.getSample(timestamp).get()).plus(new Rotation2d(orangeCamPose[0])))
            .transformBy(new Transform2d(new Pose2d(new Translation2d(), new Rotation2d(orangeCamPose[0])), Pose2d.kZero));

        robotPose = new Pose2d(robotPose.getTranslation(), new Rotation2d( gyroBuffer.getSample(timestamp).get()));

        return robotPose;
    }

    public Pose2d getBlackGlobalPose(TimeInterpolatableBuffer<Double> gyroBuffer){
        return CameraToGlobalPose2d(getBlackClosestTag(), orange_dist.get(),
            getBlackRelativeX(), getBlackRelativeY(), Timer.getTimestamp(), gyroBuffer);
    }

    public double getBlackRelativeX(){
        return blackRobotRelativeX.get();
    }

    public double getBlackRelativeY(){
        return blackRobotRelativeY.get();
    }

    public double getBlackRelativeZ(){
        return blackRobotRelativeZ.get();
    }

    public int getBlackClosestTag(){
        return (int) blackClosestTag.get();
    }

    public double getBlackTimeStamp(){
        return blackRobotRelativeX.getLastChange();
    }

    public boolean getBlackTagDetected(){
        return blackTagDetected.get();
    }

    public double getBlackFPS(){
        return blackFramerate.get();
    }

    public Pose2d getOrangeGlobalPose(TimeInterpolatableBuffer<Double> gyroBuffer){
        if(getOrangeClosestTag() >= 1 && getOrangeClosestTag() <= 22){
            return CameraToGlobalPose2d(getOrangeClosestTag(), orange_dist.get(),
                orange_tx.get(), orange_ty.get(), Timer.getTimestamp(), gyroBuffer
                );
        }else{
            return new Pose2d();
        }
    }
    
    public double getOrangeTX(){
        return orange_tx.get();
    }

    public double getOrangeTY(){
        return orange_ty.get();
    }

    public double getOrangeDist(){
        return orange_dist.get();
    }

    public int getOrangeClosestTag(){
        return (int) orangeClosestTag.get();
    }

    public double getOrangeTimeStamp(){
        return orange_tx.getLastChange();
    }

    public boolean getOrangeTagDetected(){
        return orangeTagDetected.get();
    }

    public double getOrangeFPS(){
        return orangeFramerate.get();
    }
    
}
